/* Like rb.c, but using the more standard m[row][column]
   order instead of m[column][row]. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bsd_rand.h"

#define N 512
#define T 32
#define C 0.5

char m[N][N];

static void rand_fill(char c, int n)
{
  while (n--) {
    int i, j;
    do {
      i = bsd_rand(N);
      j = bsd_rand(N);
    } while (m[j][i] != ' ');
    m[j][i] = c;
  }
}

static void init_array()
{
  memset(m, ' ', sizeof(m));
  rand_fill('r', N * N / 4);
  rand_fill('b', N * N / 4);
}

static void print_array()
{
  int j;

  for (j = 0; j < N; j++) {
    fwrite(m[j], N, 1, stdout);
    fwrite("\n", 1, 1, stdout);
  }
}

static int count_cells(int c, int ti, int tj)
{
  int i, j, cnt = 0;

  ti *= T;
  tj *= T;

  for (i = 0; i < T; i++)
    for (j = 0; j < T; j++) {
      if (m[tj + j][ti + i] == c)
        cnt++;
    }

  return cnt;
}

static int check_regions()
{
  int i, done = 0;

  #pragma omp parallel for reduction(||:done)
  for (i = 0; i < N / T; i++) {
    int j;
    for (j = 0; !done && j < N / T; j++) {
      int rc = count_cells('r', i, j);
      int bc = count_cells('b', i, j);

      if ((rc > (T * T * C))
          || (bc > (T * T * C)))
        done = 1;
    }
  }

  return done;
}

int main(int argc, char **argv)
{
  int n, cnt = 0, show = 0, red_moved, blue_moved;

  if (argc < 2) {
    fprintf(stderr, "need an argument\n");
    return 1;
  }

  show = argc > 2;
    
  n = atoi(argv[1]);
  bsd_srand(n);
  
  init_array();

  if (show)
    printf("%d\n", N);

  do {
    int i, j;

    if (show)
      print_array();

    if (check_regions())
      break;

    red_moved = 0;

    #pragma omp parallel for default(none) shared(m) private(i) reduction(||:red_moved)
    for (j = 0; j < N; j++) {
      int left_white = (m[j][0] == ' ');
      for (i = 0; i < N-1; i++) {
        if ((m[j][i] == 'r') && (m[j][i+1] == ' ')) {
          m[j][i] = ' ';
          m[j][i+1] = 'r';
          red_moved = 1;
          i++; /* don't move again */
        }
      }
      if (i < N) {
        if ((m[j][N-1] == 'r') && left_white) {
          m[j][N-1] = ' ';
          m[j][0] = 'r';
          red_moved = 1;
        }
      }
    }

    blue_moved = 0;

    #pragma omp parallel for default(none) shared(m) private(j) reduction(||:blue_moved)
    for (i = 0; i < N; i++) {
      int top_white = (m[0][i] == ' ');
      for (j = 0; j < N-1; j++) {
        if ((m[j][i] == 'b') && (m[j+1][i] == ' ')) {
          m[j][i] = ' ';
          m[j+1][i] = 'b';
          blue_moved = 1;
          j++; /* don't move again */
        }
      }
      if (j < N) {
        if ((m[N-1][i] == 'b') && top_white) {
          m[N-1][i] = ' ';
          m[0][i] = 'b';
          blue_moved = 1;
        }
      }
    }

    cnt++;
  } while (red_moved || blue_moved);

  if (!show)
    printf("%d\n", cnt);

  return 0;
}
